# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import List
from aibolit.patterns.nested_blocks.nested_blocks import NestedBlocks, BlockType


class NestedLoop:
    '''
    Returns lines in the file where
    nested FOR/IF blocks are located
    '''

    def __init__(self):
        pass

    def value(self, filename: str) -> List[int]:
        '''Return line numbers in the file where patterns are found'''
        pattern = NestedBlocks(
            2,
            [
                BlockType.WHILE,
                BlockType.FOR,
                BlockType.DO
            ]
        )
        return pattern.value(filename)



from javalang.ast import Node
from typing import List, Tuple
from aibolit.utils.ast import AST


class Lines:
    """
    Return the lines for some AST
    """
    def __init__(self, filename: str):
        self._filename = filename
        self._ast = AST(filename)
        self.value()

    def value(self) -> Tuple[Node, List[str]]:
        with open(self._filename, encoding=self._ast.encoding) as file:
            lines = file.readlines()

        return self._ast.value(), lines



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import javalang
from typing import List
import numpy as np  # type: ignore
from scipy.stats import entropy  # type: ignore

from aibolit.utils.ast import AST


class Entropy:
    def __init__(self):
        pass

    def __file_to_tokens(self, filename: str) -> List[str]:
        '''Takes path to java class file and returns tokens'''
        ast = AST(filename)
        with open(filename, encoding=ast.encoding) as file:
            tokens = javalang.tokenizer.tokenize(file.read())

        return list(map(lambda v: v.value, tokens))

    def value(self, filename: str):
        tokens = self.__file_to_tokens(filename)
        values, counts = np.unique(tokens, return_counts=True)
        return entropy(counts)



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from hashlib import md5
from pathlib import Path
from unittest import TestCase

from aibolit.config import Config
from lxml import etree

from aibolit.__main__ import list_dir, calculate_patterns_and_metrics, \
    create_xml_tree, create_text, format_converter_for_pattern


class TestRecommendPipeline(TestCase):

    def __init__(self, *args, **kwargs):
        super(TestRecommendPipeline, self).__init__(*args, **kwargs)
        self.cur_file_dir = Path(os.path.realpath(__file__)).parent
        self.config = Config.get_patterns_config()

    def __create_mock_input(self):
        patterns = [x['code'] for x in self.config['patterns']]
        item = {
            'filename': '1.java',
            'results': [
                {'pattern_code': 'P23',
                 'pattern_name': 'Some patterns name',
                 'code_lines': [1, 2, 4]
                 }
            ],
            'importances': sum([0.1 + x for x in range(len(patterns))])
        }
        another_item = {
            'filename': 'hdd/home/jardani_jovonovich/John_wick.java',
            'results': [
                {'pattern_code': 'P2',
                 'pattern_name': 'Somebody please get this man a gun',
                 'code_lines': [10, 100, 15000]},
                {'pattern_code': 'P4',
                 'pattern_name': 'New item',
                 'code_lines': [5, 6]}
            ],
            'importances': sum([0.1 + 2 * x for x in range(len(patterns))])
        }
        error_file = {
            'error_string': "Error occured",
            'filename': 'hdd/home/Error.java',
            'results': []
        }
        mock_input = [item, another_item, error_file]
        return mock_input

    def test_calculate_patterns_and_metrics(self):
        file = Path(self.cur_file_dir, 'folder/LottieImageAsset.java')
        calculate_patterns_and_metrics(file)

    def test_list_dir_no_java_files(self):
        found_files = []
        file = Path(self.cur_file_dir, 'no_java_files')
        self.assertEqual(list_dir(file, found_files), found_files)

    def test_list_dir(self):
        file = Path(self.cur_file_dir, 'folder')
        found_files = []
        list_dir(file, found_files)
        resuls = {'KeyframeParser.java', 'Metadata.java', 'LottieImageAsset.java'}
        filenames = set([Path(x).name for x in found_files])
        self.assertEqual(filenames, resuls)

    def test_xml_create_full_report(self):
        patterns = [x['code'] for x in self.config['patterns']]
        item = {
            'filename': '1.java',
            'results': [
                {'pattern_code': 'P23',
                 'pattern_name': 'Some patterns name',
                 'code_lines': [1, 2, 4]
                 }
            ],
            'importances': sum([0.1 + x for x in range(len(patterns))])
        }
        another_item = {
            'filename': 'hdd/home/jardani_jovonovich/John_wick.java',
            'results': [
                {'pattern_code': 'P2',
                 'pattern_name': 'Somebody please get this man a gun',
                 'code_lines': [10, 100, 15000]},
                {'pattern_code': 'P4',
                 'pattern_name': 'New item',
                 'code_lines': [5, 6]}
            ],
            'importances': sum([0.1 + 2 * x for x in range(len(patterns))])
        }
        error_file = {
            'error_string': "Error occured",
            'filename': 'hdd/home/Error.java',
            'results': []
        }
        mock_input = [item, another_item, error_file]
        xml_string = create_xml_tree(mock_input, full_report=True)
        md5_hash = md5(etree.tostring(xml_string))
        self.assertEqual(md5_hash.hexdigest(), '35f56275d4ba073e8d9c89b143c124da')

    def test_xml_empty_resutls(self):
        xml_string = create_xml_tree([], True)
        md5_hash = md5(etree.tostring(xml_string))
        self.assertEqual(md5_hash.hexdigest(), '7d55be99025f9d9bba410bdbd2c42cee')

    def test_text_format(self):
        mock_input = self.__create_mock_input()
        new_mock = format_converter_for_pattern(mock_input)
        text = create_text(new_mock, full_report=True)
        md5_hash = md5('\n'.join(text).encode('utf-8'))
        self.assertEqual(md5_hash.hexdigest(), 'e59a6eced350dc1320dffc2b99dcfecd')

    def test_empty_text_format(self):
        new_mock = format_converter_for_pattern([])
        text = create_text(new_mock, full_report=True)
        md5_hash = md5('\n'.join(text).encode('utf-8'))
        self.assertEqual(md5_hash.hexdigest(), 'bc22beda46ca18267a677eb32361a2aa')



import javalang
from aibolit.utils.ast import AST


class ErClass:

    def __init__(self):
        pass

    def value(self, filename: str):
        classes = ('manager',
                   'controller',
                   'router',
                   'dispatcher',
                   'printer',
                   'writer',
                   'reader',
                   'parser',
                   'generator',
                   'renderer',
                   'listener',
                   'producer',
                   'holder',
                   'interceptor')
        tree = AST(filename).value().filter(javalang.tree.ClassDeclaration)
        return [node._position.line for _, node in tree if [n for n in classes if n in node.name.lower()] != []]



import javalang
from aibolit.patterns.var_middle.var_middle import JavalangImproved


class newJavalangImproved(JavalangImproved):

    def filter(self, ntypes):
        nodes = self.tree_to_nodes()
        array = []
        for index, i in enumerate(nodes):
            if (type(i.node) in ntypes):
                if type(i.node.then_statement) in [javalang.tree.BlockStatement] and i.node.else_statement is not None:
                    for check_return in i.node.then_statement.statements:
                        if type(check_return) in [javalang.tree.ReturnStatement]:
                            array.append(nodes[index].line)
        return array


class CountIfReturn:
    '''
    Returns lines with if statement which has also return statement
    and other conditions with else.
    '''
    def __init__(self):
        pass

    def value(self, filename: str):
        ''''''
        tree = newJavalangImproved(filename)
        if_decls = tree.filter([javalang.tree.IfStatement])

        return if_decls



from typing import List

import javalang

from aibolit.types_decl import LineNumber
from aibolit.utils.ast import AST


class ClassicSetter:

    def __init__(self):
        pass

    def value(self, filename: str) -> List[LineNumber]:
        lst: List[LineNumber] = []
        tree = AST(filename).value().filter(javalang.tree.MethodDeclaration)
        for path, node in tree:
            if (node.return_type is None) and ('set' in node.name[:3]):
                if (isinstance(node.body, list)) and len(node.body) < 2:
                    for statement in node.body:
                        if isinstance(statement, javalang.tree.StatementExpression):
                            if isinstance(statement.expression, javalang.tree.Assignment):
                                expression = statement.expression.expressionl
                                if isinstance(expression, javalang.tree.This):
                                    if statement.expression.type == '=':
                                        if expression.selectors[0].member.lower() == node.name.lower()[3:]:
                                            lst.append(node._position.line)
                                    else:
                                        break
                                else:
                                    break
                            else:
                                break
        return lst




class VarSiblings:
    def __init__(self):
        pass

    def value(self, filename: str):
        # @todo #141:30min: Implement var_siblings pattern.
        #  This pattern must return a tuple containing the line numbers of the two siblings. After implementing that
        #  activate the tests in test_car_sibling.py
        return []



import javalang
from aibolit.utils.ast import AST


class ProtectedMethod:

    def __init__(self):
        pass

    def value(self, filename: str):
        return [
            node.position.line
            for _, node in AST(filename).value().filter(javalang.tree.MethodDeclaration)
            if all(elem in node.modifiers for elem in ['protected'])
        ]



#!/usr/bin/env python
# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""The main entry point. Invoke as `aibolit' or `python -m aibolit'.
"""

import argparse
import concurrent.futures
import multiprocessing
import sys
from os import scandir
from typing import List

from lxml import etree  # type: ignore
import numpy as np  # type: ignore
from aibolit import __version__
from aibolit.config import Config
from aibolit.ml_pipeline.ml_pipeline import train_process, collect_dataset
import os
from pathlib import Path
import pickle
from aibolit.model.model import TwoFoldRankingModel, Dataset  # type: ignore  # noqa: F401
from sys import stdout

dir_path = os.path.dirname(os.path.realpath(__file__))


def list_dir(path, files):
    dir_list = []
    for entry in scandir(path):
        if entry.is_dir():
            dir_list.append(entry.path)
            dir_list.extend(list_dir(entry.path, files))
        elif entry.is_file() and entry.path.endswith('.java'):
            files.append(entry.path)
    return dir_list


def predict(input_params, model, args):
    features_order = model.features_conf['features_order']
    # load model
    input = [input_params[i] for i in features_order]
    th = float(args.threshold) or 1.0
    preds, importances = model.predict(np.array(input), th=th)

    return {features_order[int(x)]: int(x) for x in preds.tolist()[0]}, importances


def run_parse_args(commands_dict):
    parser = argparse.ArgumentParser(
        description='Find the pattern which has the largest impact on readability',
        usage='''
        aibolit <command> [<args>]

        You can run 1 command:
        train          Train model
        recommend      Recommend pattern
        ''')

    parser.add_argument('command', help='Subcommand to run')
    parser.add_argument(
        '--version', action='version',
        version='%(prog)s {version}'.format(version=__version__)
    )
    # parse_args defaults to [1:] for args, but you need to
    # exclude the rest of the args too, or validation will fail
    args = parser.parse_args(sys.argv[1:2])
    if args.command not in commands_dict:
        parser.print_help()
        exit(1)

    return commands_dict[args.command]()


def train():
    parser = argparse.ArgumentParser(
        description='Collect dataset and train model'
    )
    parser.add_argument(
        '--java_folder',
        help='full path to directory',
        default=False,
        required=False
    )
    parser.add_argument(
        '--max_classes',
        type=lambda v: sys.maxsize if v == '' else int(v),
        required=False,
        default=None
    )
    parser.add_argument(
        '--skip_collect_dataset',
        required=False,
        default=False,
        action='store_true'
    )
    parser.add_argument(
        '--dataset_file',
        required=False,
        default=False
    )
    args = parser.parse_args(sys.argv[2:])
    if not args.skip_collect_dataset:
        collect_dataset(args)
    train_process()


def __count_value(value_dict, input_params, code_lines_dict, java_file: str, is_metric=False):
    """
    Count value for input dict

    :param value_dict: Pattern item or Metric item from CONFIG
    :param input_params: list with calculated patterns/metrics
    :param code_lines_dict: list with found code lines of patterns/metrics
    :param java_file: full path for java file
    :is_metric: is item metric?
    :return: None, it has side-effect
    """
    acronym = value_dict['code']
    try:
        val = value_dict['make']().value(java_file)
        if not is_metric:
            input_params[acronym] = len(val)
            code_lines_dict['lines_' + acronym] = val
        else:
            input_params[acronym] = val
    except Exception:
        exc_type, exc_value, exc_tb = sys.exc_info()
        raise Exception("Can't count {} metric: {}".format(
            acronym,
            str(exc_value))
        )


def calculate_patterns_and_metrics(file):
    code_lines_dict = input_params = {}  # type: ignore
    error_string = None
    try:
        config = Config.get_patterns_config()
        for pattern in config['patterns']:
            if pattern['code'] in config['patterns_exclude']:
                continue
            __count_value(pattern, input_params, code_lines_dict, file)

        for metric in config['metrics']:
            if metric['code'] in config['metrics_exclude']:
                continue
            __count_value(metric, input_params, code_lines_dict, file, is_metric=True)
    except Exception as ex:
        error_string = str(ex)
        input_params = []  # type: ignore

    return input_params, code_lines_dict, error_string


def inference(
        input_params: List[int],
        code_lines_dict,
        args):
    """
    Find a pattern which has the largest impact on target

    :param input_params: list if calculated patterns/metrics
    :param code_lines_dict: list with found code lines of patterns/metrics
    :param file: filename

    :return: list of results with code_lies for each pattern and its name
    """
    model_path = args.model_file
    do_full_report = args.full
    results = []
    importances = [-1]
    if input_params:
        if not model_path:
            model_path = Config.folder_model_data()
        with open(model_path, 'rb') as fid:
            model = pickle.load(fid)
        sorted_result, importances = predict(input_params, model, args)
        patterns_list = model.features_conf['patterns_only']
        for iter, (key, val) in enumerate(sorted_result.items()):
            if key in patterns_list:
                pattern_code = key
                code_lines = code_lines_dict.get('lines_' + key)
                # We show only patterns with positive importance
                if code_lines and val > 1.00000e-20:
                    if code_lines:
                        pattern_name = \
                            [x['name'] for x in Config.get_patterns_config()['patterns']
                             if x['code'] == pattern_code][0]
                        results.append(
                            {'code_lines': code_lines,
                             'pattern_code': pattern_code,
                             'pattern_name': pattern_name
                             })
                    if not do_full_report:
                        break
    else:
        return results, sum(importances)

    return results, sum(importances)


def run_recommend_for_file(file: str, args):
    """
    Calculate patterns and metrics, pass values to model and recommend pattern to change
    :param file: file to analyze
    :param args: different command line arguments
    :return: dict with code lines, filename and pattern name
    """
    java_file = str(Path(os.getcwd(), file))
    input_params, code_lines_dict, error_string = calculate_patterns_and_metrics(java_file)
    results_list, importances = inference(input_params, code_lines_dict, args)

    return {
        'filename': file,
        'results': results_list,
        'error_string': error_string,
        'importances': importances
    }


def create_xml_tree(results, full_report):
    """
    Creates xml from output of `recommend` function
    :param results: output of `recommend` function
    :return: xml string
    """
    importances_for_all_classes = []
    top = etree.Element('report')
    importances_for_all_classes_tag = etree.SubElement(top, 'score')
    files = etree.SubElement(top, 'files')
    if not full_report:
        files.addprevious(etree.Comment('Show pattern with the largest contribution to Cognitive Complexity'))
    else:
        files.addprevious(etree.Comment('Show all patterns'))
    for result_for_file in results:
        file = etree.SubElement(files, 'file')
        filename = result_for_file.get('filename')
        name = etree.SubElement(file, 'path')
        output_string_tag = etree.SubElement(file, 'summary')
        name.text = filename
        results = result_for_file.get('results')
        errors_string = result_for_file.get('error_string')
        if not results and not errors_string:
            output_string = 'Your code is perfect in aibolit\'s opinion'
            output_string_tag.text = output_string
        elif not results and errors_string:
            output_string = 'Error when calculating patterns: {}'.format(str(errors_string))
            output_string_tag.text = output_string
        else:
            output_string = 'Some issues found'
            output_string_tag.text = output_string
            importances_sum_tag = etree.SubElement(file, 'score')
            importances_value_per_class = result_for_file['importances']
            importances_sum_tag.text = str(importances_value_per_class)
            importances_for_all_classes.append(importances_value_per_class)
            patterns_tag = etree.SubElement(file, 'patterns')
            for pattern in result_for_file['results']:
                if pattern.get('pattern_code'):
                    pattern_item = etree.SubElement(patterns_tag, 'pattern')
                    pattern_name_str = pattern.get('pattern_name')
                    details = etree.SubElement(pattern_item, 'details')
                    details.text = pattern_name_str or ''
                    pattern_item.attrib['code'] = pattern.get('pattern_code')
                    code_lines_items = pattern.get('code_lines')
                    if code_lines_items:
                        code_lines_lst_tree_node = etree.SubElement(pattern_item, 'lines')
                        for code_line in code_lines_items:
                            code_line_elem = etree.SubElement(code_lines_lst_tree_node, 'number')
                            code_line_elem.text = str(code_line)
    if importances_for_all_classes:
        importances_for_all_classes_tag.text = str(np.mean(importances_for_all_classes))

    return top


def get_exit_code(results):
    """
    Analyzed recommendation results and generate exit_code for pipeline
    """

    files_analyzed = len(results)
    errors_number = 0
    perfect_code_number = 0
    for result_for_file in results:
        results = result_for_file.get('results')
        errors_string = result_for_file.get('error_string')
        if not results and not errors_string:
            perfect_code_number += 1
        elif not results and errors_string:
            errors_number += 1

    if errors_number == files_analyzed:
        # we have errors everywhere
        exit_code = 2
    elif perfect_code_number == files_analyzed:
        # everything is good
        exit_code = 0
    else:
        # we have some recommendation
        exit_code = 1

    return exit_code


def create_text(results, full_report):
    importances_for_all_classes = []
    buffer = []
    if not full_report:
        buffer.append('Show pattern with the largest contribution to Cognitive Complexity')
    else:
        buffer.append('Show all patterns')
    for result_for_file in results:
        filename = result_for_file.get('filename')
        buffer.append('Filename {}: '.format(filename))
        results = result_for_file.get('results')
        errors_string = result_for_file.get('error_string')
        if not results and not errors_string:
            output_string = 'Your code is perfect in aibolit\'s opinion'
            buffer.append(output_string)
        elif not results and errors_string:
            output_string = 'Error when calculating patterns: {}'.format(str(errors_string))
            buffer.append(output_string)
        else:
            output_string = 'Some issues found'
            score = result_for_file['importances']
            importances_for_all_classes.append(score)
            buffer.append('Score for file: {}'.format(score))
            buffer.append(output_string)
            for pattern_item in result_for_file['results']:
                code = pattern_item.get('pattern_code')
                if code:
                    pattern_name_str = pattern_item.get('pattern_name')
                    buffer.append('line {}: {} ({})'.format(pattern_item.get('code_line'), pattern_name_str, code))
    if importances_for_all_classes:
        buffer.append('Total score: {}'.format(np.mean(importances_for_all_classes)))

    return buffer


def recommend():
    """Run recommendation pipeline."""

    parser = argparse.ArgumentParser(
        description='Get recommendations for Java code',
        usage='''
        aibolit recommend < --folder | --filenames > [--output] [--model_file] [--threshold] [--full] [--format]
        ''')

    group_exclusive = parser.add_mutually_exclusive_group(required=True)

    group_exclusive.add_argument(
        '--folder',
        help='path to Java files',
        default=False
    )
    group_exclusive.add_argument(
        '--filenames',
        help='list of Java files',
        nargs="*",
        default=False
    )
    parser.add_argument(
        '--model_file',
        help='''file where pretrained model is located, the default path is located
        in site-packages and is installed with aibolit automatically''',
        default=False
    )

    parser.add_argument(
        '--threshold',
        help='threshold for predict',
        default=False
    )
    parser.add_argument(
        '--full',
        help='show all recommendations instead of the best one',
        default=False,
        action='store_true'
    )
    parser.add_argument(
        '--format',
        default='text',
        help='text (by default) or xml. Usage: --format=xml'
    )

    args = parser.parse_args(sys.argv[2:])

    if args.threshold:
        print('Threshold for model has been set to {}'.format(args.threshold))

    if args.filenames:
        files = args.filenames
    elif args.folder:
        files = []
        list_dir(args.folder, files)

    results = list(run_thread(files, args))

    if args.format:
        new_results = format_converter_for_pattern(results)
        if args.format == 'text':
            text = create_text(new_results, args)
            print('\n'.join(text))
        elif args.format == 'xml':
            root = create_xml_tree(results, args.full)
            tree = root.getroottree()
            tree.write(stdout.buffer, pretty_print=True)
        else:
            raise Exception('Unknown format')

    exit_code = get_exit_code(results)
    return exit_code


def format_converter_for_pattern(results):
    """Reformat data where data are sorted by patterns importance
    (it is already sorted in the input).
    Then lines are sorted in ascending order."""

    def flatten(l):
        return [item for sublist in l for item in sublist]

    for file in results:
        items = file.get('results')
        if items:
            new_items = flatten([
                [{'pattern_code': x['pattern_code'],
                  'pattern_name': x['pattern_name'],
                  'code_line': line,
                  } for line in sorted(x['code_lines'])] for x in items
            ])
            file['results'] = new_items

    return results


def version():
    """
    Parses arguments and shows current version of program.
    """

    parser = argparse.ArgumentParser(
        description='Show version')
    parser.add_argument(
        '--version',
    )
    print('%(prog)s {version}'.format(version=__version__))


def run_thread(files, args):
    """
    Parallel patterns/metrics calculation
    :param files: list of java files to analyze

    """
    with concurrent.futures.ThreadPoolExecutor(max_workers=multiprocessing.cpu_count()) as executor:
        future_results = [executor.submit(run_recommend_for_file, file, args) for file in files]
        concurrent.futures.wait(future_results)
        for future in future_results:
            yield future.result()


def main():
    try:
        commands = {
            'train': lambda: train(),
            'recommend': lambda: recommend(),
            'version': lambda: version(),
        }
        exit_code = run_parse_args(commands)
    except Exception:
        import traceback
        traceback.print_exc()
        sys.exit(2)
    else:
        sys.exit(exit_code)


if __name__ == '__main__':
    main()



import javalang
from aibolit.utils.ast import AST


class SuperMethod:

    def __init__(self):
        pass

    def value(self, filename: str):
        """
        Iterates over functions and finds super.func() calls.
        Javalang doesn't have code line for super.func() call,
        that's why we can only count the first match of a call inside some function.
        It has MULTIPLE MATCHES if we call super.func() inside a ANONYMOUS CLASS.
        :param filename:
        :return: Lines of code
        """
        results = []
        ast = AST(filename)
        tree = ast.value()

        with open(filename, encoding=ast.encoding) as file:
            text_lines = file.readlines()
        for _, method_decl_node in tree.filter(javalang.tree.MethodDeclaration):
            code_line = method_decl_node.position.line
            for _, super_method_inv in method_decl_node.filter(javalang.tree.SuperMethodInvocation):
                str_to_find = 'super.{method_name}('.format(
                    method_name=super_method_inv.member).strip()
                for iter, line in enumerate(text_lines[code_line - 1:]):
                    string_strip = line.strip().replace('\n', '').replace('\t', '')
                    if string_strip.find(str_to_find) > -1:
                        results.append(code_line + iter)
                        break
        return results

    def __traverse(self, tree, results):
        descendants = tree.children
        for children in descendants:
            if isinstance(children, tuple) or isinstance(children, list):
                for item in children:
                    if isinstance(item, javalang.tree.SuperMethodInvocation):
                        results.append([item.member])
                    else:
                        self.__traverse(item, results)
        return results



import javalang

from aibolit.utils.ast import AST


class MultipleWhile:

    def __init__(self):
        pass

    def value(self, filename: str):
        """
        Travers over AST tree and finds function with sequential while statement
        :param filename:
        :return:
        List of LineNumber of methods which have sequential while statements
        """

        res = []
        for _, method_node in AST(filename).value().filter(javalang.tree.MethodDeclaration):
            if len(list(method_node.filter(javalang.tree.WhileStatement))) > 1:
                res.append(method_node.position.line)

        return res



from typing import List
from javalang.tree import ClassDeclaration
from aibolit.types_decl import LineNumber
from aibolit.utils.ast import AST


class NonFinalClass:

    def __init__(self):
        pass

    def value(self, filename: str) -> List[LineNumber]:
        tree = AST(filename).value()
        classes = tree.filter(ClassDeclaration)
        return [
            node.position.line for _, node in classes
            if len([v for v in ['final', 'abstract'] if v in node.modifiers]) == 0
        ]



import javalang
from aibolit.utils.ast import AST


class ImplementsMultiFinder:

    def __init__(self):
        pass

    def value(self, filename: str):
        tree = AST(filename).value().filter(javalang.tree.ClassDeclaration)
        return [node._position.line for _, node in tree if node.implements and (len(node.implements) > 1)]



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import javalang
from aibolit.utils.ast import AST


class EmptyRethrow:

    def __init__(self):
        pass

    def value(self, filename):
        tree = AST(filename).value()
        total_code_lines = set()
        for _, method_node in tree.filter(javalang.tree.MethodDeclaration):
            for _, try_node in method_node.filter(javalang.tree.TryStatement):
                for _, throw_node in try_node.filter(javalang.tree.ThrowStatement):
                    if try_node.catches:
                        catch_classes = [x.parameter.name for x in try_node.catches]
                        mem_ref = throw_node.children[1]
                        if isinstance(mem_ref, javalang.tree.ClassCreator):
                            continue
                        else:
                            if hasattr(mem_ref, 'member') and mem_ref.member in catch_classes:
                                total_code_lines.add(mem_ref.position.line)
        return sorted(total_code_lines)



from javalang.parse import parse
from javalang.tree import CompilationUnit
import cchardet  # type: ignore


class AST:
    """
    Returns the AST for some java file
    """

    def __init__(self, filename: str):
        self._filename = filename
        with open(self._filename, 'rb') as file:
            bin_data = file.read()

        self.encoding = cchardet.detect(bin_data)['encoding']

    def value(self) -> CompilationUnit:

        with open(self._filename, 'r', encoding=self.encoding) as file:  # type: ignore
            return parse(file.read())



from typing import List
from javalang.ast import Node
from javalang.tree import (
    CompilationUnit,
    AssertStatement,
    ClassDeclaration
)

from aibolit.types_decl import LineNumber
from aibolit.utils.ast import AST

_TEST_CLASS_SUFFIX = 'Test'


class AssertInCode(object):
    def value(self, filename: str) -> List[LineNumber]:
        tree: CompilationUnit = AST(filename).value()

        return self.__traverse_node(tree)

    def __traverse_node(self, tree: CompilationUnit) -> List[LineNumber]:
        lines: List[LineNumber] = list()

        for path, node in tree.filter(AssertStatement):
            lines.append(node.position.line)
        return lines


def _within_test_class(path) -> bool:
    class_declaration = next(filter(_is_class_declaration, path[::2]))

    return class_declaration.name.endswith(_TEST_CLASS_SUFFIX)


def _is_class_declaration(node: Node) -> bool:
    return isinstance(node, ClassDeclaration)



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import unittest
from aibolit.metrics.ncss.ncss import NCSSMetric


class TestNCSSMetric(unittest.TestCase):
    def testZeroScore(self):
        file = 'test/metrics/ncss/Empty.java'
        metric = NCSSMetric()
        res = metric.value(file)
        self.assertEqual(res, 0)

    def testLowScore(self):
        file = 'test/metrics/ncss/GraalSDK.java'
        metric = NCSSMetric()
        res = metric.value(file)
        self.assertEqual(res, 2)

    def testMediumScore(self):
        file = 'test/metrics/ncss/WorkflowRunActionRepetitionDefinitionInner.java'
        metric = NCSSMetric()
        res = metric.value(file)
        self.assertEqual(res, 73)

    def testHighScore(self):
        file = 'test/metrics/ncss/YarnConfiguration.java'
        metric = NCSSMetric()
        res = metric.value(file)
        self.assertEqual(res, 1301)

    def testBasicExample(self):
        file = 'test/metrics/ncss/BasicExample.java'
        metric = NCSSMetric()
        res = metric.value(file)
        self.assertEqual(res, 12)

    def testAnotherScore(self):
        file = 'test/metrics/ncss/WebAppsImpl.java'
        metric = NCSSMetric()
        res = metric.value(file)
        self.assertEqual(res, 1429)



from typing import List

import numpy as np
import pandas as pd
from catboost import CatBoost
from sklearn.base import BaseEstimator
from sklearn.preprocessing import StandardScaler

from aibolit.config import Config


class Dataset:

    def __init__(self, only_patterns: List[str]):
        self.input = None
        self.target = None
        self.do_rename_columns = False
        self.only_patterns = only_patterns

    def preprocess_file(
            self,
            scale_ncss=True,
            scale=False,
            **kwargs):

        print('reading dataset from {}'.format(Config.dataset_file()))
        df = pd.read_csv(Config.dataset_file())
        df = df[~df["filename"].str.lower().str.contains("test")]
        config = Config.get_patterns_config()
        if self.do_rename_columns:
            p_codes = \
                [x['code'] for x in config['patterns']] \
                + ['lines' + x['code'] for x in config['patterns']]
            m_codes = [x['code'] for x in config['metrics']]
            keys = p_codes + m_codes
            vals = \
                [x['name'] for x in config['patterns']] \
                + ['lines' + x['name'] for x in config['patterns']] \
                + [x['name'] for x in config['metrics']]

            replace_dict = dict(zip(keys, vals))
            df = df.rename(replace_dict)
            df.columns = vals
            print('Columns renamed:' + df.head())

        df = df.dropna().drop_duplicates(subset=df.columns.difference(['filename']))
        df = df[(df.ncss > 20) & (df.ncss < 100) & (df.npath_method_avg < 100000.00)].copy().reset_index()

        df.drop('filename', axis=1, inplace=True)
        df.drop('index', axis=1, inplace=True)
        self.target = np.array(df[['M4']].values[:, 0], dtype=np.float64)
        if scale_ncss:
            new = pd.DataFrame(
                df[self.only_patterns].values / df['M2'].values.reshape((-1, 1)),
                columns=self.only_patterns
            )
            self.target /= df['M2'].values.reshape(-1)
        else:
            new = df[self.only_patterns].copy()
        if scale:
            self.input = pd.DataFrame(StandardScaler().fit_transform(new.values), columns=new.columns,
                                      index=new.index).values
        else:
            self.input = new.values

        self.feature_order = list(new.columns)


class TwoFoldRankingModel(BaseEstimator):

    def __init__(self):
        self.do_rename_columns = False
        self.model = None
        self.features_conf = None

    def fit(self, X, y, display=False):
        """
        Args:
            X: np.array with shape (number of snippets, number of patterns) or
                (number of patterns, ).
            y: np.array with shape (number of snippets,), array of snippets'
                complexity metric values
            display: bool, to output info about training or not
        """
        model = CatBoost()

        grid = {'learning_rate': [0.03, 0.1],
                'depth': [4, 6, 10],
                'l2_leaf_reg': [1, 3, 5, 7, 9]}

        model.grid_search(
            grid,
            X=X,
            y=y,
            verbose=display,
        )

        self.model = model
        self.model.fit(X, y.ravel(), logging_level='Silent')

    def __get_pairs(self, item, th: float):
        def sigmoid(x):
            return 1 / (1 + np.exp(-x))

        pattern_importances = item * self.model.feature_importances_
        # mask discards not significant patterns
        th_mask = (sigmoid(pattern_importances) <= th) + 0
        pattern_importances = pattern_importances * th_mask
        order = np.arange(self.model.feature_importances_.size)
        return (pattern_importances, order)

    def __vstack_arrays(self, res):
        return np.vstack(res).T

    def predict(self, X, quantity_func='log', th=1.0):
        """
        Args:
            X: np.array with shape (number of snippets, number of patterns) or
                (number of patterns, ).
            quantity_func: str, type of function that will be applied to
                number of occurrences.
            th (float): Sensitivity of algorithm to recommend.
                0 - ignore all recomendations
                1 - use all recommendations

        Returns:
            ranked: np.array with shape (number of snippets, number of patterns)
                of sorted patterns in non-increasing order for each snippet of
                code.
        """

        if X.ndim != 1:
            raise Exception("Only input with size 1, N is allowed")
        else:
            X = X.copy()
            X = np.expand_dims(X, axis=0)

        ranked = []
        quantity_funcs = {
            'log': lambda x: np.log1p(x) / np.log(10),
            'exp': lambda x: np.exp(x + 1),
            'linear': lambda x: x,
        }

        for snippet in X:
            try:
                item = quantity_funcs[quantity_func](snippet)
                pairs = self.__vstack_arrays(self.__get_pairs(item, th))
                pairs = pairs[pairs[:, 0].argsort()]
                ranked.append(pairs[:, 1].T.tolist()[::-1])
            except Exception:
                raise Exception("Unknown func")

        return (np.array(ranked), pairs[:, 0].T.tolist()[::-1])




from typing import List

from javalang.tree import FieldDeclaration

from aibolit.types_decl import LineNumber
from aibolit.utils.ast import AST


class NonFinalAttribute:

    def __init__(self):
        pass

    def value(self, filename: str) -> List[LineNumber]:
        tree = AST(filename).value().filter(FieldDeclaration)

        return [node.position.line for path, node in tree if 'final' not in node.modifiers]



import os
import unittest
from aibolit.patterns.var_siblings.var_siblings import VarSiblings
from pathlib import Path


class TestVarSiblings(unittest.TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent

    @unittest.skip("Not implemented")
    def test_find_simple_var_siblings(self):
        self.assertEqual(
            [(3, 4)], VarSiblings(Path(self.dir_path, 'SimpleVarSiblings.java')).value()
        )

    @unittest.skip("Not implemented")
    def test_find_alternate_var_siblings(self):
        self.assertEqual(
            [(3, 5)], VarSiblings(Path(self.dir_path, 'AlternateVarSiblings.java')).value()
        )



from javalang.tree import IfStatement, SwitchStatement, ForStatement, WhileStatement
from javalang.tree import DoStatement, CatchClause, BreakStatement, ContinueStatement
from javalang.tree import TernaryExpression, BinaryOperation, MethodDeclaration, MethodInvocation
from aibolit.utils.ast import AST
import javalang
from typing import List, Any, Type

increment_for: List[Type] = [
    IfStatement,
    SwitchStatement,
    ForStatement,
    WhileStatement,
    DoStatement,
    CatchClause,
    BreakStatement,
    ContinueStatement,
    TernaryExpression,
    BinaryOperation,
    MethodInvocation,
]

nested_for: List[Type] = [
    IfStatement,
    SwitchStatement,
    ForStatement,
    WhileStatement,
    DoStatement,
    CatchClause,
]

logical_operators: List[str] = ['&', '&&', '^', '|', '||']


class CognitiveComplexity:
    '''
    beta version of Cognitive Complexity
    '''
    def __init__(self):
        self.complexity = 0
        self.method_name = None

    def _traverse_childs(self, block: Any, nested_level: int) -> None:

        for each_child in block.children:
            self._get_complexity(each_child, nested_level)

    def _check_if_statement(self, expr: IfStatement, nested_level: int) -> None:
        '''function to work with IfStatement block'''
        self._get_complexity(expr.condition, 0)
        if expr.then_statement is not None:
            self.complexity += nested_level + 1
            self._get_complexity(expr.then_statement, nested_level + 1)

        if expr.else_statement is not None:
            if isinstance(expr.else_statement, IfStatement):
                self.complexity -= nested_level
                self._check_if_statement(expr.else_statement, nested_level)
            else:
                self.complexity += 1
                self._get_complexity(expr.else_statement, nested_level + 1)

    def _increment_logical_operators(self, block: BinaryOperation, prev_operator: str) -> None:
        for each_block in [block.operandr, block.operandl]:

            if isinstance(each_block, BinaryOperation) and prev_operator in logical_operators:
                if prev_operator != each_block.operator:
                    self.complexity += 1

                elif isinstance(each_block.operandr, BinaryOperation):
                    self.complexity += 1

                self._increment_logical_operators(each_block, each_block.operator)

    def _is_recursion_call(self, block: MethodInvocation) -> bool:
        if self.method_name == block.member:
            return True
        return False

    def _nested_methods(self, block: MethodDeclaration, nested_level: int) -> None:
        original_name = self.method_name
        self.method_name = block.name
        self._get_complexity(block, nested_level + 1)
        self.method_name = original_name

    def _get_complexity(self, block: Any, nested_level: int) -> None:
        block_arr = block if isinstance(block, List) else [block]

        for each_block in block_arr:
            if hasattr(each_block, 'children'):

                if type(each_block) == MethodDeclaration and each_block.name != self.method_name:
                    self._nested_methods(each_block, nested_level)

                elif isinstance(each_block, IfStatement):
                    self._check_if_statement(each_block, nested_level)

                elif type(each_block) in increment_for and type(each_block) in nested_for:
                    self.complexity += 1 + nested_level
                    self._traverse_childs(each_block, nested_level + 1)

                elif type(each_block) in increment_for and type(each_block) not in nested_for:
                    if isinstance(each_block, BinaryOperation):
                        if each_block.operator in logical_operators:
                            self.complexity += 1
                            self._increment_logical_operators(each_block, each_block.operator)

                    elif isinstance(each_block, MethodInvocation):
                        is_recursion = self._is_recursion_call(each_block)
                        self.complexity += is_recursion

                    else:
                        self.complexity += 1
                        self._traverse_childs(each_block, nested_level)

                else:
                    self._traverse_childs(each_block, nested_level)

    def value(self, filename: str) -> int:

        tree = AST(filename).value()
        for _, class_body in tree.filter(javalang.tree.ClassDeclaration):
            for each_object in class_body.body:
                if isinstance(each_object, MethodDeclaration):

                    # memorize the name for detecting recursion call
                    self.method_name = each_object.name
                    self._get_complexity(each_object, 0)

        final_value, self.complexity = self.complexity, 0
        return final_value



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from collections import defaultdict
from aibolit.utils.ast import AST

import javalang


class ReturnNull:

    def __init__(self):
        pass

    def value(self, filename: str):
        """
        Travers over AST tree and finds pattern
        :param filename:
        """
        tree = AST(filename).value()
        chain_lst = defaultdict(int)
        for _, method_node in tree.filter(javalang.tree.MethodDeclaration):
            for _, return_node in method_node.filter(javalang.tree.ReturnStatement):
                return_literal = return_node.children[1]
                if isinstance(return_literal, javalang.tree.Literal) and return_literal.value == 'null':
                    chain_lst[method_node.name] = return_literal.position.line or return_node.position.line
                elif isinstance(return_literal, javalang.tree.TernaryExpression):
                    chain_lst[method_node.name] = return_node.position.line

        filtered_dict = list(filter(lambda elem: elem > 0, chain_lst.values()))
        return filtered_dict



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from aibolit.utils.ast import AST
import javalang


class ArrayAsArgument:

    def __init__(self):
        pass

    def value(self, filename):
        tree = AST(filename).value().filter(javalang.tree.MethodDeclaration)
        ret = []
        for path, node in tree:
            if len(node.parameters) > 0:
                if any(len(parameter.type.dimensions) > 0 for parameter in node.parameters):
                    ret += [node.position.line]
        return ret



import os
from unittest import TestCase
from pathlib import Path
from aibolit.utils.ast import AST


class TestAST(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent

    def test_open_file(self):
        ast = AST(Path(self.dir_path, 'ConditionalExpressionCheck.java'))
        ast.value()
        self.assertEqual(ast.encoding, 'UTF-8')

    def test_utf(self):
        ast = AST(Path(self.dir_path, 'ExceptionDemo.java'))
        ast.value()
        self.assertEqual(ast.encoding, 'GB18030')



import javalang
from aibolit.utils.ast import AST
from typing import List


class SendNull:

    def __init__(self):
        pass

    def value(self, filename: str) -> List[int]:
        lst: List[int] = []
        tree = AST(filename).value()

        invocation_tree = tree.filter(javalang.tree.Invocation)
        arg_list = [x for _, x in invocation_tree]

        for argument in arg_list:
            ternary_list = argument.filter(javalang.tree.TernaryExpression)
            for path, expr in ternary_list:
                if isinstance(expr.if_false, javalang.tree.Literal) and expr.if_false.value == 'null':
                    lst.append(argument._position.line)
                if isinstance(expr.if_true, javalang.tree.Literal) and expr.if_true.value == 'null':
                    lst.append(argument._position.line)

        for path, node in tree:
            try:
                for argument in node.arguments:
                    if isinstance(argument, javalang.tree.Literal) and argument.value == "null" and \
                            argument._position.line not in lst:
                        lst.append(argument._position.line)
            except (AttributeError, TypeError):
                pass
        lst = sorted(lst)
        return lst



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from typing import List
import sys
from aibolit.types_decl import LineNumber
from aibolit.patterns.var_middle.var_middle import JavalangImproved
from javalang.tree import BinaryOperation, ThrowStatement, IfStatement


class JoinedValidation:
    """
    Pattern which matches joined validations: validations (if with a single throw inside) which condition
    contains more than condition joined with OR
    """

    def __init__(self):
        pass

    def value(self, filename) -> List[LineNumber]:
        """
        Returns the line number of joined validations found in file.
        """

        def c1(node):
            return type(node.node.condition) == BinaryOperation

        def c2(node):
            return node.node.condition.operator == '||'

        def c3(node):
            return type(node.node.then_statement) == ThrowStatement

        def c4(node):
            return hasattr(node.node.then_statement, "statements") and len(node.node.then_statement.statements) == 1

        def c5(node):
            return type(node.node.then_statement.statements[0]) == ThrowStatement

        sys.setrecursionlimit(5000)
        return [
            node.line
            for node in JavalangImproved(filename).filter([IfStatement])
            if (c1(node) and c2(node)) and (c3(node) or (c4(node) and c5(node)))
        ]



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import javalang
from aibolit.utils.ast import AST


class ForceTypeCastingFinder:

    def __process_node(self, node):
        line = node.position.line if hasattr(node, 'position') and node.position is not None else None
        qualifier = node.qualifier if hasattr(node, 'qualifier') else None
        member = node.member if hasattr(node, 'member') else None
        name = node.name if hasattr(node, 'name') else None
        return {
            "line": line,
            "name": qualifier or member or name,
            "ntype": type(node)
        }

    def __tree_to_list(self, tree: javalang.tree.CompilationUnit):
        '''Convert AST tree to list of object'''
        items = [self.__process_node(node) for path, node in tree if node is not None]

        # fill missed line numbers
        last_line_number = None

        for item in items:
            if (item['line']) is not None:
                last_line_number = item['line']
                continue
            item['line'] = last_line_number

        return items

    def value(self, filename: str):
        ''''''
        tree = AST(filename).value()
        list_tree = self.__tree_to_list(tree)
        num_str = []
        for node in list_tree:
            if node['ntype'] == javalang.tree.Cast:
                k = int(node['line'])
                if k not in num_str:
                    num_str.append(k)
        return num_str



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from aibolit.utils.ast import AST


class NCSSMetric():
    def __init__(self):
        pass

    def value(self, filename: str):
        if len(filename) == 0:
            raise ValueError('Empty file for analysis')

        tree = AST(filename).value()

        metric = 0
        for path, node in tree:
            node_type = str(type(node))
            if 'Statement' in node_type:
                metric += 1
            elif 'VariableDeclarator' == node_type:
                metric += 1
            elif 'Assignment' == node_type:
                metric += 1
            elif 'Declaration' in node_type and 'LocalVariableDeclaration' not in node_type \
                 and 'PackageDeclaration' not in node_type:
                metric += 1

        return metric
